import warnings

import numpy as np

from stockroom import parser
from stockroom.utils import LazyLoader

torch = LazyLoader('torch', globals(), 'torch')
tf = LazyLoader('tf', globals(), 'tensorflow')


class Model:
    """
    Model class utilizes hangar columns to store pieces of a model and use hangar
    metadata to store the information required to collate it back to a model. Currently,
    it supports ``keras.Model`` and ``torch.nn.Module`` models. ModelStore instance,
    on :meth:`stockroom.storages.Model.save_weights` creates few columns (one column for
    each data type) to store the weights and create one column specifically to store the
    shape of each layer. This shape column is needed because the weights of each layer
    would be flattened before saving. This is essential since handling variable shapes
    and variable ranks are more complex than flattening and reshaping-back the weights.

    Examples
    --------
    >>> import torch
    >>> import tensorflow as tf
    >>> torch_model = torch.Sequential(...)
    >>> stock.model['torch_model'] = torch_model.state_dict()
    >>> tf_model = tf.Keras.Sequential()
    >>> tf_model.add(tf.layers.Dense(64, activation='relu'))
    >>> stock.model['tf_model'] = tf_model.get_weights()

    But if you can make it easy by calling special functions that knows how to fetch
    weights from the model or how to put weights back to model. Checkout :meth:`Model.save_weights`
    & :meth:`Model.load_weights` for more details
    """
    def __init__(self, accessor):
        self.accessor = accessor

    def __setitem__(self, name, weights):
        if isinstance(weights, dict):
            layers = weights.keys()
            weights = [x.numpy() for x in weights.values()]
            library = 'torch'
            library_version = torch.__version__
        elif isinstance(weights, list):
            library = 'tf'
            layers = None
            library_version = tf.__version__
        else:
            raise TypeError("Unknown type. Weights has to be a dict or list")
        longest = max([len(x.reshape(-1)) for x in weights])
        dtypes = [w.dtype.name for w in weights]
        writer = self.accessor
        metakey = parser.model_metakey(name)
        if metakey not in writer.columns.keys():
            metacol = writer.add_str_column(metakey)
        else:
            metacol = writer[metakey]
        metacol['library'] = library
        metacol['libraryVersion'] = library_version
        metacol['longest'] = str(longest)
        metacol['dtypes'] = parser.stringify(dtypes)
        metacol['numLayers'] = str(len(weights))
        metacol['layers'] = parser.stringify(layers)

        # ---------- Create ndarray columns if doesn't exist -----------------
        shapeKey = parser.model_shapekey(name, str(longest))
        if shapeKey not in writer.columns.keys():
            shape_typ = np.array(1).dtype  # C long = int32 in win64; int64 elsewhere
            writer.add_ndarray_column(shapeKey, 10, shape_typ, variable_shape=True)
        for i, w in enumerate(weights):
            modelKey = parser.modelkey(name, str(longest), dtypes[i])
            if modelKey not in writer.columns.keys():
                writer.add_ndarray_column(
                    modelKey, longest, np.dtype(dtypes[i]), variable_shape=True)
        # ---------------------------------------------------------

        shape_col = writer.columns[shapeKey]
        for i, w in enumerate(weights):
            model_col = writer.columns[parser.modelkey(name, longest, dtypes[i])]
            model_col[i] = w.reshape(-1)

            if w.shape:
                shape_col[i] = np.array(w.shape)
            else:
                # C long = int32 in win64; int64 elsewhere
                shape_typ = np.array(1).dtype
                shape_col[i] = np.array(()).astype(shape_typ)

    def __getitem__(self, name):
        reader = self.accessor
        try:
            metakey = parser.model_metakey(name)
            metacol = reader.columns[metakey]
        except KeyError:
            raise KeyError(f"Model with key {name} not found")
        library = metacol['library']
        library_version = metacol['libraryVersion']
        longest = int(metacol['longest'])
        dtypes = parser.destringify(metacol['dtypes'])
        num_layers = int(metacol['numLayers'])
        layers = parser.destringify(metacol['layers'])

        shapeKey = parser.model_shapekey(name, longest)
        shape_col = reader.columns[shapeKey]
        weights = []
        for i in range(num_layers):
            modelKey = parser.modelkey(name, longest, dtypes[i])
            col = reader.columns[modelKey]
            w = col[i].reshape(np.array(shape_col[i]))
            weights.append(w)
        if library == 'torch':
            if torch.__version__ != library_version:
                warnings.warn(f"PyTorch version used while storing the model "
                              f"({library_version}) is not same as the one installed "
                              f"in the current environment. i.e {torch.__version__}")
            return {layers[i]: torch.from_numpy(weights[i]) for i in range(num_layers)}

        else:
            if tf.__version__ != library_version:
                warnings.warn(f"Tensorflow version used while storing the model "
                              f"({library_version}) is not same as the one installed "
                              f"in the current environment. i.e {tf.__version__}")
            return weights



from pathlib import Path
import inspect
from stockroom.external.importer.base import BaseImporter
from stockroom.external.importer import torchvision_importers


def is_valid(x):
    return inspect.isclass(x) and issubclass(x, BaseImporter) and x != BaseImporter


importers_dict = {'torchvision': {cls.name: cls
                                  for _, cls in inspect.getmembers(torchvision_importers, is_valid)}}


def get_importers(source: str, download_dir: Path):
    try:
        package, dataset = source.split('.')
    except ValueError:
        raise RuntimeError(f"Could not parse the source string '{source}'")

    try:
        return (importers_dict[package][dataset](download_dir, train=True),
                importers_dict[package][dataset](download_dir, train=False))
    except KeyError:
        raise RuntimeError("Could not fetch the dataset you were looking for. "
                           "Create a request for new importers")



import pytest


def test_basic(writer_stock):
    writer_stock.experiment['lr'] = 0.01
    writer_stock.experiment['epochs'] = 500
    writer_stock.experiment['optimizer'] = 'adam'
    writer_stock.commit('Saved lr')

    assert writer_stock.experiment['lr'] == 0.01
    assert writer_stock.experiment['optimizer'] == 'adam'
    assert writer_stock.experiment['epochs'] == 500


def test_save_string(writer_stock):
    with pytest.raises(TypeError):
        writer_stock.experiment['wrongdata'] = bytes('hi')



from pathlib import Path
import shutil
import pytest
import hangar
import numpy as np
from stockroom import StockRoom, keeper


@pytest.fixture()
def managed_tmpdir(monkeypatch, tmp_path):
    monkeypatch.setitem(hangar.constants.LMDB_SETTINGS, 'map_size', 2_000_000)
    monkeypatch.setitem(hangar.constants.LMDB_SETTINGS, 'map_size', 2_000_000)
    monkeypatch.setattr(hangar.backends.hdf5_00, 'COLLECTION_COUNT', 10)
    monkeypatch.setattr(hangar.backends.hdf5_00, 'COLLECTION_SIZE', 50)
    monkeypatch.setattr(hangar.backends.hdf5_01, 'COLLECTION_COUNT', 10)
    monkeypatch.setattr(hangar.backends.hdf5_01, 'COLLECTION_SIZE', 50)
    monkeypatch.setattr(hangar.backends.numpy_10, 'COLLECTION_SIZE', 50)
    yield tmp_path
    shutil.rmtree(tmp_path)


@pytest.fixture()
def repo(monkeypatch, managed_tmpdir):
    cwd = Path(managed_tmpdir)
    monkeypatch.setattr(Path, 'cwd', lambda: cwd)
    cwd.joinpath(".git").mkdir()
    cwd.joinpath(".gitignore").touch()
    keeper.init_repo('s', 'a@b.c', overwrite=True)
    yield None


@pytest.fixture()
def repo_with_col(repo):
    repo = hangar.Repository(Path.cwd())
    co = repo.checkout(write=True)
    arr = np.arange(20).reshape(4, 5)
    co.add_ndarray_column('ndcol', prototype=arr)
    co.commit('init column')
    co.close()
    yield None
    repo._env._close_environments()


@pytest.fixture()
def writer_stock(repo_with_col):
    stock_obj = StockRoom(write=True)
    yield stock_obj
    stock_obj._repo._env._close_environments()



from pathlib import Path
import pytest
import os
import stockroom.utils as utils


def test_get_stock_root(repo):
    cwd = Path.cwd()
    root = utils.get_stock_root(cwd)
    assert root == cwd

    os.rmdir(cwd/'.git')
    root = utils.get_stock_root(cwd)
    assert root == cwd

    root = utils.get_stock_root(cwd/'dummy/path')
    assert root == cwd

    path = Path('/down/the/rabbit/hole')
    with pytest.raises(RuntimeError):
        utils.get_stock_root(path)



import abc


class BaseImporter(abc.ABC):
    name = None

    @abc.abstractmethod
    def column_names(self):
        """
        Return the names for each data entitiy. For instance, if each data sample returns
        an image and a label, this function could return `(image, label)` so that
        stockroom creates an image column for image data and label column for label data
        """

    @abc.abstractmethod
    def shapes(self):
        """
        Return the shape of the data sample(s). Return a tuple of shapes if one data
        sample has more than one data entity. For instance, if the data sample is
        image and label, this function need to return `(shape of image, shape of label)`.
        If the shape is variable, return the maximum value of shape in each dimension.
        """
        pass

    @abc.abstractmethod
    def dtypes(self):
        """
        Return the datatype of the data sample. The data must be homogenous across
        dataset. If one data sample has more than one data entity, return a tuple
        of dtypes. For instance, if the data sample is image and label, then this
        function should return `(dtype of image, dtype of label)`.
        """
        pass

    @abc.abstractmethod
    def variability_status(self):
        """
        Return True/False depending on the shape of the data sample across the
        dataset is variable or not. If the data sample contains more than one data
        entity, this function should return tuple of boolean values. For instance,
        if the data sample is image and label, then this function should return
        `(is_variable(image), is_variable(label))` where `is_variable` is a custom
        function to check the variability across the dataset.
        """
        pass

    @abc.abstractmethod
    def __iter__(self):
        """
        Yield data sample from the dataset. If the data sample contains more than one
        data entity, this function should return a tuple. For instance, if the data
        sample is image and label, then this function should return `(image, label)`.

        Note that the keys for each data sample would be integers incrementing by one
        from zero to len(dataset). In the current implementation, setting custom keys
        is not allowed but might enable in the future releases
        """
        pass

    @abc.abstractmethod
    def __len__(self):
        """
        Return the length of the dataset. Simlar to `len(pytorch_dataset)`
        """
        pass



import pytest
import numpy as np


def test_save_data(writer_stock):
    stock = writer_stock
    arr = np.arange(20).reshape(4, 5)
    col = stock.data['ndcol']
    col[1] = arr
    stock.commit("added data")
    assert np.allclose(stock.data['ndcol', 1], arr)
    del stock


def test_fetch_non_existing_column(writer_stock):
    arr = np.arange(20).reshape(4, 5)
    with pytest.raises(KeyError):
        writer_stock.data['wrongcol']


def test_save_to_different_typed_column(writer_stock):
    arr = np.arange(20).reshape(4, 5).astype(np.float)
    col = writer_stock.data['ndcol']
    with pytest.raises(ValueError):
        col[1] = arr


def test_fetch_non_existing_sample_key(writer_stock):
    arr = np.arange(20).reshape(4, 5)
    col = writer_stock.data['ndcol']
    col[1] = arr
    with pytest.raises(KeyError):
        col[2]



from pathlib import Path
from contextlib import ExitStack

import click
from click_didyoumean import DYMGroup
from hangar import Repository

from stockroom.keeper import init_repo
from stockroom.core import StockRoom
from stockroom import __version__
from stockroom import external


@click.group(no_args_is_help=True, add_help_option=True,
             invoke_without_command=True, cls=DYMGroup)
@click.version_option(version=__version__, help='display current stockroom version')
def stock():
    """
    The `stock` CLI provides a git-like experience (whenever possible) to interact with
    the stock repository. It also means that the *current working directory* is where
    the stock repository would exist (like git 😊 ).
    """
    pass


@stock.command()
@click.option('--username', prompt='Username', help='Username of the user')
@click.option('--email', prompt='User Email', help='Email address of the user')
@click.option('--overwrite', is_flag=True, default=False,
              help='overwrite a repository if it exists at the current path')
def init(username, email, overwrite):
    """
    Init stockroom repository. This will create a .hangar directory and a `head.stock`
    file in your `cwd`. `stock init` would be triggered implicitly if you are making
    a stock repository by using `stock import` but in all other case, you'd need to
    initialize a stock repository to start operating on it with the python APIs
    """
    try:
        init_repo(username, email, overwrite)
    except RuntimeError as e:
        raise click.ClickException(e)  # type: ignore


@stock.command()
@click.option('--message', '-m', multiple=True,
              help=('The commit message. If multiple arguments are provided, '
                    'each of them gets converted into a new line'))
def commit(message):
    """
    It does a stock commit. Stock commit consists of two actions

    1. Make a hangar commit
    2. Update the `head.stock` file (git will track this file if you are using git)
    """
    if len(message) < 1:
        raise click.ClickException(ValueError("Require commit message"))
    # TODO: There should be a way to share the write enabled checkout if user need to
    #  commit let's say when he has a writer checkout open in jupyter notebook
    stock_obj = StockRoom(write=True)
    msg = '\n'.join(message)
    click.echo('Commit message:\n' + msg)
    try:
        digest = stock_obj.commit(message)
    except (FileNotFoundError, RuntimeError) as e:
        raise click.ClickException(e)  # type: ignore
    click.echo(f'Commit Successful. Digest: {digest}')


@stock.command(name='import')
@click.argument('dataset_name')
@click.option('--download-dir', '-d', default=Path.home(), type=click.Path(),
              help=('If you have the dataset downloaded in a non-default path or want '
                    'to download it to a non-default path, pass it here'))
def import_data(dataset_name, download_dir):
    """
    Downloads and add a pytorch dataset (from torchvision, torchtext or torchaudio)
    to StockRoom. It creates the repo if it doesn't exist and loads the dataset
    into a repo for you
    """
    try:
        stock_obj = StockRoom(write=True)
    except RuntimeError:
        repo = Repository('.', exists=False)
        if not repo.initialized:
            raise RuntimeError("Repository is not initialized. Check `stock init --help` "
                               "details about how to initialize a repository")
        else:
            raise
    # TODO: use the auto-column-creation logic in stockroom later
    co = stock_obj.accessor
    importers = external.get_importers(dataset_name, download_dir)
    total_len = sum([len(importer) for importer in importers])
    with click.progressbar(label='Adding data to StockRoom', length=total_len) as bar:
        for importer in importers:
            column_names = importer.column_names()
            dtypes = importer.dtypes()
            shapes = importer.shapes()
            for colname, dtype, shape in zip(column_names, dtypes, shapes):
                if colname not in co.keys():
                    # TODO: this assuming importer always return a numpy flat array
                    co.add_ndarray_column(colname, dtype=dtype, shape=shape)
            columns = [co[name] for name in column_names]
            with ExitStack() as stack:
                for col in columns:
                    stack.enter_context(col)
                for i, data in enumerate(importer):
                    bar.update(1)
                    for col, dt in zip(columns, data):
                        # TODO: use the keys from importer
                        col[i] = dt
    stock_obj.commit(f'Data from {dataset_name} added through stock import')
    stock_obj.close()
    click.echo(f'The {dataset_name} dataset has been added to StockRoom')



import types
import importlib
from pathlib import Path


def get_stock_root(path: Path) -> Path:
    """
    Traverse from given path up till root of the system to figure out the root of the
    stock repo. A stock repo must be hangar repo, a git repo and must have a head.stock
    file. The head.stock file has the information required for stockroom to manage
    checkout, branching etc and it is git tracked.

    Parameters
    ----------
    path : Path
        path from which stock root check starts

    Returns
    -------
    Path
        Location of root of stock repo
    """
    while True:
        stock_exist = path.joinpath('head.stock').exists()
        if stock_exist:
            hangar_exist = path.joinpath('.hangar').exists()
            git_exist = path.joinpath('.git').exists()
            if not hangar_exist and not git_exist:
                raise RuntimeError("Stock root should be the root of git and"
                                   "hangar repository")
            return path
        if path == path.parent:  # system root check
            raise RuntimeError("Could not find stock root. Are you in a "
                               "stock repository")
        path = path.parent


def get_current_head(root: Path) -> str:
    """
    Reads the stock file and return the commit hash if found

    Parameters
    ----------
    root : Path
        The stock root path

    Returns
    -------
    str
        commit hash if found. Empty string other wises
    """
    with open(root/'head.stock', 'r') as f:
        commit = f.read()
        return commit if commit else ''


def set_current_head(root: Path, commit: str):
    """
    Write a commit hash to the stock file.

    Parameters
    ----------
    root : Path
        The stock root path
    commit : str
        Commit hash that will be written to the stock file
    """
    with open(root/'head.stock', 'w+') as f:
        f.write(commit)


class LazyLoader(types.ModuleType):
    """
    Lazily import a module, mainly to avoid pulling in large dependencies
    """

    def __init__(self, local_name, parent_module_globals, name):
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals
        super(LazyLoader, self).__init__(name)

    def _load(self):
        """Load the module and insert it into the parent's globals.

        Import the target module and insert it into the parent's namespace
        Update this object's dict so that if someone keeps a reference to the
        LazyLoader, lookups are efficient (__getattr__ is only called on
        lookups that fail).
        """
        module = importlib.import_module(self.__name__)
        self._parent_module_globals[self._local_name] = module
        self.__dict__.update(module.__dict__)
        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)




from stockroom import StockRoom
from random import randint
import numpy as np
import pytest


class TestSameProcess:

    def test_opening_two_instances(self, writer_stock):
        with pytest.raises(PermissionError):
            StockRoom(write=True)
        arr = np.arange(20).reshape(4, 5)
        oldarr = arr * randint(1, 100)
        col1 = writer_stock.data['ndcol']
        col1[1] = oldarr
        writer_stock.commit('added data')

        stock2 = StockRoom()
        col2 = stock2.data['ndcol']

        assert np.allclose(col2[1], oldarr)
        stock2._repo._env._close_environments()



class Data:
    """
    Data storage is essentially a wrapper over hangar's column API which let stockroom
    handles the checkout scope. The instance creation is not something user would
    directly do here. Instead, a created instance will be available at :class:`stockroom.StockRoom`

    Note
    ----
    Each ``__getitem__`` or ``__setitem__`` call will open & close a hangar checkout.
    Unlike other storages, this is a crucial information for data storage because both
    reading and writing of data happens quite frequently in a pipeline unlike saving or
    retrieving model or parameters or metrics. So for optimizing, this you could make the
    data read/write inside the context manager :meth:`stockroom.StockRoom.optimize`

    Examples
    --------
    >>> stock = StockRoom()
    >>> stock.data['column1']['sample1'] = np.arange(20).reshape(5, 4)
    >>> sample = stock.data['column1']['sample5']

    Inside context manager

    >>> with stock.run():
    ...     sample = stock.data['coloumn1']['sample1']
    """
    def __init__(self, accessor):
        self.accessor = accessor

    def __getattr__(self, item):
        # TODO: guard for non-allowed
        return getattr(self.accessor, item)

    def __setitem__(self, key, value):
        self.accessor[key] = value

    def __getitem__(self, key):
        return self.accessor[key]



from pathlib import Path
import warnings
from hangar import Repository


def init_repo(name=None, email=None, overwrite=False):
    """ init hangar repo, create stock file and add details to .gitignore """
    if not Path.cwd().joinpath('.git').exists():
        warnings.warn("initializing stock repository in a directory which is not a "
                      "git repository. Some features won't work", UserWarning)
    repo = Repository(Path.cwd(), exists=False)
    if not overwrite and repo.initialized:
        commit_hash = repo.log(return_contents=True)['head']
        print(f'Hangar Repo already exists at {repo.path}. '
              f'Initializing it as stock repository')
    else:
        if name is None or email is None:
            raise ValueError("Both ``name`` and ``email`` cannot be None")
        commit_hash = ''
        repo.init(user_name=name, user_email=email, remove_old=overwrite)
    # closing the environment for avoiding issues in windows
    repo._env._close_environments()

    stock_file = Path.cwd()/'head.stock'
    if stock_file.exists():
        warnings.warn("Trying to initialize an already initialized stock repository. "
                      "No action taken", UserWarning)
    else:
        with open(str(stock_file), 'w+') as f:
            f.write(commit_hash)
        print("Stock file created")

    gitignore = Path.cwd()/'.gitignore'
    with open(str(gitignore), 'a+') as f:
        f.seek(0)
        if '.hangar' not in f.read():
            f.write('\n# hangar artifacts\n.hangar\n')



import pytest
import numpy as np
from copy import deepcopy

import torch


def get_model():
    torch_model = torch.nn.Sequential(
        torch.nn.Linear(2, 3),
        torch.nn.ReLU(),
        torch.nn.Linear(3, 1))
    return torch_model


# TODO: Ingore warning has no effect
@pytest.mark.filterwarnings('ignore:the imp module is deprecated:DeprecationWarning')
def test_saving_model(writer_stock):
    model = get_model()
    old_weights = model.state_dict()
    writer_stock.model['model'] = old_weights
    writer_stock.commit('adding model')

    model = get_model()
    tmp_weights = deepcopy(model.state_dict())
    model.load_state_dict(writer_stock.model['model'])
    new_weights = deepcopy(model.state_dict())
    iterator = old_weights.keys() if isinstance(old_weights, dict) else range(len(old_weights))
    for k in iterator:
        assert np.allclose(old_weights[k], new_weights[k])
        assert not tmp_weights[k].sum() == 0.0 \
            or np.allclose(tmp_weights[k], new_weights[k])


def test_saving_two_models(writer_stock):
    model1 = get_model()
    model2 = get_model()
    weights1 = model1.state_dict()
    weights2 = model2.state_dict()
    writer_stock.model[1] = weights1
    writer_stock.model[2] = weights2
    writer_stock.commit('adding two models')
    ret_weights1 = writer_stock.model[1]
    ret_weights2 = writer_stock.model[2]
    model1.load_state_dict(ret_weights1)
    model2.load_state_dict(ret_weights2)
    iterator = ret_weights1.keys() if isinstance(ret_weights1, dict) else range(len(ret_weights1))
    for k in iterator:
        assert np.allclose(ret_weights1[k], weights1[k])
        assert np.allclose(ret_weights2[k], weights2[k])


def test_load_with_different_library_version(writer_stock, monkeypatch):
    model = get_model()
    writer_stock.model['model'] = model.state_dict()
    writer_stock.commit("adding model")
    with monkeypatch.context() as m:
        m.setattr(torch, '__version__', '0.111')
        with pytest.warns(UserWarning) as warning_rec:
            model.load_state_dict(writer_stock.model['model'])
        assert len(warning_rec) == 1
        assert "version used while storing the model" in warning_rec[0].message.args[0]


def test_unknown_model_type(writer_stock):
    with pytest.raises(TypeError):
        writer_stock.model['invalid'] = set()


def test_load_nonexisting_key(writer_stock):
    model = get_model()
    writer_stock.model['model'] = model.state_dict()
    writer_stock.commit("adding model")
    with pytest.raises(KeyError) as error:
        writer_stock.model['wrongname']
    assert 'Model with key wrongname not found' == error.value.args[0]



from pathlib import Path
from stockroom import init_repo
import pytest
import hangar
import hangar.checkout


class TestInit:
    @staticmethod
    @pytest.fixture(autouse=False)
    def repo_path(monkeypatch, managed_tmpdir):
        path = Path(managed_tmpdir)
        monkeypatch.setattr(Path, 'cwd', lambda: path)
        return path

    @staticmethod
    @pytest.fixture(autouse=False)
    def cwd(repo_path):
        repo_path.joinpath('.git').mkdir()
        return repo_path

    def test_init(self, repo_path):
        cwd = repo_path
        cwd.joinpath(".git").mkdir()
        with pytest.raises(ValueError):
            init_repo(overwrite=True)
        with pytest.raises(ValueError):
            init_repo(name='some', overwrite=True)
        with pytest.raises(ValueError):
            init_repo(email='some', overwrite=True)
        init_repo('s', 'a@b.c', overwrite=True)
        with open(cwd.joinpath('.gitignore')) as f:
            assert '\n.hangar\n' in f.read()
        assert cwd.joinpath('.hangar').exists()
        assert cwd.joinpath('head.stock').exists()

    def test_stock_init_on_existing_hangar_repo(self, cwd):
        repo = hangar.Repository(cwd, exists=False)
        repo.init('a', 'a@b.c')
        repo._env._close_environments()
        assert not cwd.joinpath('head.stock').exists()
        init_repo()
        assert cwd.joinpath('head.stock').exists()
        with open(cwd.joinpath('.gitignore')) as f:
            assert '\n.hangar\n' in f.read()


class TestCommit:
    def test_basic(self, writer_stock):
        writer_stock.experiment['key1'] = 'value'
        writer_stock.commit('generic data')
        assert writer_stock.experiment['key1'] == 'value'
        writer_stock.experiment['key2'] = 'value2'

    def test_commit_hash(self, writer_stock):
        writer_stock.experiment['key1'] = 'value'
        writer_stock.commit('generic data')
        with open(writer_stock.stockroot/'head.stock') as f:
            digest1 = f.read()
        writer_stock.experiment['key2'] = 'value2'
        writer_stock.commit('generic data 2')
        with open(writer_stock.stockroot/'head.stock') as f:
            digest2 = f.read()
        log = writer_stock._repo.log(return_contents=True)
        log['order'].pop()  # removing the digest from conftest.py
        assert log['order'] == [digest2, digest1]


def test_hangar_checkout_from_stock_obj(writer_stock):
    co = writer_stock.accessor
    assert isinstance(co, hangar.checkout.WriterCheckout)
    assert hasattr(co, '_writer_lock')

    co.close()
    with pytest.raises(PermissionError):
        writer_stock.experiment['key1'] = 'value'



from pathlib import Path

import pytest
from click.testing import CliRunner

import stockroom.cli as cli
from stockroom import StockRoom


def test_version():
    import stockroom
    runner = CliRunner()
    res = runner.invoke(cli.stock, ['--version'])
    assert res.exit_code == 0
    assert res.stdout == f'stock, version {stockroom.__version__}\n'


def test_init_repo():
    runner = CliRunner()
    with runner.isolated_filesystem():
        with pytest.raises(RuntimeError):
            StockRoom()
        res = runner.invoke(cli.init, ['--username', 'test', '--email', 'test@foo.com'])
        assert res.exit_code == 0
        StockRoom()


def test_commit(repo_with_col):
    runner = CliRunner()
    stock = StockRoom(write=True)
    stock.experiment['key'] = 'value'
    stock.close()
    res = runner.invoke(cli.commit, [])
    assert 'Error: Require commit message\n' in res.stdout
    res = runner.invoke(cli.commit, ['-m', 'test commit'])
    assert res.exit_code == 0
    assert 'Commit message:\ntest commit' in res.stdout
    assert 'Commit Successful. Digest' in res.stdout
    stock._repo._env._close_environments()



try:
    from torchvision import datasets
except ModuleNotFoundError:
    pass
import numpy as np

from stockroom.external.importer.base import BaseImporter


class TorchvisionCommon(BaseImporter):

    def __init__(self, dataset, train):
        self.dataset = dataset
        self.split = 'train' if train else 'test'
        self.sample_img, self.sample_label = self._process_data(*self.dataset[0])

    def column_names(self):
        return f'{self.name}-{self.split}-image', f'{self.name}-{self.split}-label'

    def shapes(self):
        return self.sample_img.shape, self.sample_label.shape

    def dtypes(self):
        return self.sample_img.dtype, self.sample_label.dtype

    @staticmethod
    def _process_data(img, lbl):
        # TODO: memory copy
        img = np.ascontiguousarray(np.transpose(np.array(img), (2, 0, 1)))
        img = img.astype(np.float32) / 255
        lbl = np.array(lbl)
        return img, lbl

    def __iter__(self):
        for img, label in self.dataset:
            yield self._process_data(img, label)

    def variability_status(self):
        return False

    def __len__(self):
        return len(self.dataset)


class Cifar10(TorchvisionCommon):
    name = 'cifar10'

    def __init__(self, root, train=True):
        dataset = datasets.CIFAR10(root=root, train=train, download=True)
        super().__init__(dataset, train)


class Mnist(TorchvisionCommon):
    name = 'mnist'

    def __init__(self, root, train=True):
        dataset = datasets.MNIST(root=root, train=train, download=True)
        super().__init__(dataset, train)


class FashionMnist(TorchvisionCommon):
    name = 'fashion_mnist'

    def __init__(self, root, train=True):
        dataset = datasets.FashionMNIST(root=root, train=train, download=True)
        super().__init__(dataset, train)



from typing import Union
from pathlib import Path
from contextlib import contextmanager, ExitStack
import logging
import time

from hangar import Repository
from stockroom.storages import Model, Data, Experiment
from stockroom.utils import get_stock_root, get_current_head, set_current_head

logger = logging.getLogger(__name__)  # TODO make the formatter and stuffs like that


class StockRoom:
    """
    This class is the only user entrypoint of stockroom that interacts with an existing
    stock repository i.e. all the repository interaction a user would do will have to go
    through an object of this class. Also, stockroom comes with three different storages

    1. Model: Weights of models built with ``keras.Model`` or ``torch.nn``
    2. Data: Dataset as numpy arrays/tensors
    3. Experiment: Information related to an experiment such as metrics, parameters etc

    An object of this class holds an object to these three storages each has a dictionary
    style access machinery

    Parameters
    ----------
    path : Union[str, Path, None]
        Path the to the stock repository. If `None`, it traverse up from `pwd` till it
        finds the stock root (stock root is the location where `head.stock` file is
        located and ideally will have `.git` folder as well

    Note
    ----
    By default (if no path is provided while initializing :class:`StockRoom`), it checks
    for the stock root. A stock root is a directory that is

    1. a git repository (has .git folder)
    2. a hangar repository (has .hangar folder)
    3. a stock repository (has head.stock file)

    If you'd like to skip these checks and just use stockroom (for example: if you are a
    hangar user and use stockroom just for storing models in your hangar repository, it
    doesn't need to be a stock repository and hence can skip these checks), provide the
    path to the repository explicitly. The rationale here is, if you provide the path, we
    trust you that you know what you doing on that path
    """
    def __init__(self, path: Union[str, Path] = None, write: bool = False):
        # TODO: context manager on Object creation
        # TODO: make force_release_writer_lock easier with stockroom?
        self.path = Path(path) if path else get_stock_root(Path.cwd())
        self._repo = Repository(self.path)
        self.head = get_current_head(self.path)  # TODO: should this be None if writer enabled
        if write:
            self.accessor = self._repo.checkout(write=True)
        else:
            if not self.head:
                self.accessor = None
            else:
                self.accessor = self._repo.checkout(commit=self.head).__enter__()

        if self.accessor:
            self.model = Model(self.accessor)
            self.data = Data(self.accessor)
            self.experiment = Experiment(self.accessor)
        else:
            self.model = None
            self.data = None
            self.experiment = None

    @contextmanager
    def run(self, autocommit=True, commit_msg=f'Auto-committing at {time.time()}'):
        # TODO: check whether opening in multiple context managers has any problem or not
        with ExitStack() as stack:
            stack.enter_context(self.accessor)
            yield
        if autocommit and self.accessor.diff.status() != 'CLEAN':
            self.accessor.commit(commit_msg)

    def update_head(self):
        if self._repo.writer_lock_held:
            logger.info("Write enabled checkouts will always be on the latest head "
                        "(staging). Doing nothing")
            return
        self.head = get_current_head(self.path)
        self.accessor.__exit__()
        self.accessor.close()
        self.accessor = self._repo.checkout(commit=self.head).__enter__()

    def close(self):
        self.accessor.close()

    @property
    def stockroot(self) -> Path:
        """
        Returns the root of stock repository
        """
        return self.path

    def commit(self, message: str, update_head=True) -> str:
        """
        Make a stock commit. A stock commit is a hangar commit plus writing the commit
        hash to the stock file. This function opens the stock checkout in write mode and
        close after the commit. Which means, no other write operations should be running
        while stock commit is in progress
        """
        digest = self.accessor.commit(message)
        set_current_head(self.stockroot, digest)
        if update_head:
            self.update_head()
        return digest



#!/usr/bin/env python

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

requirements = ['Click>=7.0', 'click_didyoumean', 'hangar>=0.5.0']

setup(
    author="Sherin Thomas",
    author_email='sherin@tensorwerk.com',
    python_requires='>=3.6',
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Database',
        'Topic :: Scientific/Engineering',
        'Topic :: Utilities',
    ],
    description="Platform to version model, data, parameters alongside git versioned source code",
    project_urls={
        'Documentation': 'https://stockroom.readthedocs.io',
        'Issue Tracker': 'https://github.com/tensorwerk/stockroom/issues'},
    entry_points={'console_scripts': ['stock=stockroom.cli:stock']},
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + '\n\n',
    long_description_content_type = 'text/markdown',
    include_package_data=True,
    keywords='stockroom',
    name='stockroom',
    packages=find_packages(include=['stockroom', 'stockroom.*']),
    url='https://github.com/tensorwerk/stockroom',
    version='0.2.2',
    zip_safe=False,
)



from stockroom import parser


class Experiment:
    """
    Tag store, as the name suggests, is to store tags related to an experiment. Ideally/
    eventually this store information on commit level and would not pass it down the
    commit history tree. But currently the internal implementation of hangar doesn't
    allow that and hence we store the information on metadata store in hangar. It
    currently takes `int`, `float` & `str` data types and convert it to a string which
    is the only data type supported by hangar metadata. But :class:`Tag` stores the type
    of the data in another metadata "column" which will be uesd while pulling the data
    back from the Tag store.

    Examples
    --------
    >>> stock.experiment['epochs'] = 1000
    >>> stock.experiment['lr'] = 0.0001
    >>> stock.experiment['optimizer'] = 'adam'
    """
    def __init__(self, accessor):
        self.typecaster = {'int': int, 'float': float, 'str': str}
        self.accessor = accessor
        self.tagkey = parser.tagkey()
        self.tagtypekey = parser.tag_typekey()

    def __setitem__(self, key, value):
        writer = self.accessor
        if isinstance(value, int):
            value_type = 'int'
        elif isinstance(value, float):
            value_type = 'float'
        elif isinstance(value, str):
            value_type = 'str'
        else:
            raise TypeError("Tag store can accept only ``int``, ``float`` or ``str``")
        if self.tagkey not in writer.columns.keys():
            writer.add_str_column(self.tagkey)
        if self.tagtypekey not in writer.columns.keys():
            writer.add_str_column(self.tagtypekey)
        writer[self.tagkey][key] = str(value)
        writer[self.tagtypekey][key] = value_type

    def __getitem__(self, key):
        reader = self.accessor
        try:
            value = reader[self.tagkey, key]
            value_type = reader[self.tagtypekey, key]
        except KeyError:
            raise KeyError(f"Data not found with key {key}")
        try:
            return self.typecaster[value_type](value)
        except KeyError:
            raise KeyError(f"Data tampering suspected. Could not "
                           f"read the data type {value_type}")



from stockroom.core import StockRoom
from stockroom.keeper import init_repo
__all__ = ['StockRoom', 'init_repo', '__version__', 'make_torch_dataset']


try:
    from hangar.dataset import make_torch_dataset
except ModuleNotFoundError:
    from hangar import make_torch_dataset


__version__ = '0.2.2'
